#!/usr/bin/env python
#
# Lara Maia <dev@lara.monster> 2015 ~ 2023
#
# The Steam Tools NG is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Steam Tools NG is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.
#
import asyncio
import binascii
import logging
from typing import Any, Optional

import aiohttp
from gi.repository import Gtk, Gdk

from stlib import login
from stlib.login import AuthCodeType
from . import utils
from .. import i18n, config, core

log = logging.getLogger(__name__)
_ = i18n.get_translation


# noinspection PyUnusedLocal
class LoginWindow(utils.PopupWindowBase):
    def __init__(
            self,
            parent_window: Gtk.Window,
            application: Gtk.Application,
            mobile_login: bool = True,
    ) -> None:
        super().__init__(parent_window, application)
        self.mobile_login = mobile_login
        self.has_user_data = False

        self.login_button = utils.AsyncButton()
        self.login_button.set_label(_("Log-in"))
        self.login_button.connect("clicked", self.on_login_button_clicked)
        self.header_bar.pack_end(self.login_button)
        self.set_title(_('Login'))

        self.status = utils.SimpleStatus()
        self.content_grid.attach(self.status, 0, 0, 1, 1)

        self.user_details_section = utils.Section("login")
        self.content_grid.attach(self.user_details_section, 0, 1, 1, 1)

        self.username_item = self.user_details_section.new_item("account_name", _("Username:"), Gtk.Entry, 0, 0)

        self.__password_item = self.user_details_section.new_item("_password", _("Password:"), Gtk.Entry, 0, 1)
        self.__password_item.set_visibility(False)
        self.__password_item.set_invisible_char('*')

        self.auth_code_item = self.user_details_section.new_item("_auth_code", _("Code:"), Gtk.Entry, 0, 2)

        self.save_password_item = self.user_details_section.new_item(
            "_savepwd",
            _("Save Password:"),
            Gtk.CheckButton,
            0, 5,
        )
        self.save_password_item.set_active(True)

        self.advanced_login = utils.ClickableLabel()
        self.advanced_login.set_markup(utils.markup(_("Advanced Login"), font_size='x-small', color='blue'))
        self.advanced_login.set_halign(Gtk.Align.END)
        self.advanced_login.set_margin_end(10)
        self.advanced_login.set_margin_bottom(10)
        self.advanced_login.connect("clicked", self.on_advanced_login_clicked)
        self.content_grid.attach(self.advanced_login, 0, 2, 1, 1)

        self.advanced_login_section = utils.Section("login")
        self.content_grid.attach(self.advanced_login_section, 0, 3, 1, 1)

        self.identity_secret_item = self.advanced_login_section.new_item(
            'identity_secret',
            _("Identity Secret:"),
            Gtk.Entry,
            0, 0,
        )

        self.shared_secret_item = self.advanced_login_section.new_item(
            'shared_secret',
            _("Shared Secret:"),
            Gtk.Entry,
            0, 1,
        )

        self.connect('destroy', self.on_quit)
        self.connect('close-request', self.on_quit)

        self.auth_code_item.set_visible(False)
        self.advanced_login_section.set_visible(False)
        self.check_login_availability()

    @property
    def username(self) -> str:
        return self.username_item.get_text()

    @property
    def __password(self) -> str:
        return self.__password_item.get_text()

    def set_password(self, encrypted_password: str) -> None:
        try:
            __password = core.utils.decode_password(encrypted_password)
            self.__password_item.set_text(__password)
        except (binascii.Error, UnicodeError, TypeError):
            log.warning(_("Password decode failed. Trying RAW."))
            self.__password_item.set_text(encrypted_password)

    @property
    def auth_code(self) -> str:
        return self.auth_code_item.get_text()

    @property
    def shared_secret(self) -> str:
        return self.shared_secret_item.get_text()

    @property
    def identity_secret(self) -> str:
        return self.identity_secret_item.get_text()

    def check_login_availability(self) -> None:
        if not self.username or not self.__password:
            self.login_button.set_sensitive(False)
        else:
            self.login_button.set_sensitive(True)

    def on_quit(self, *args: Any, **kwargs: Any) -> None:
        self.application.main_window.statusbar.set_warning(
            'steamguard',
            _("Login cancelled! Modules will not work correctly!"),
        )
        self.destroy()

    def on_key_released_event(
            self,
            controller: Gtk.EventControllerKey,
            keyval: int,
            keycode: int,
            state: Gdk.ModifierType
    ) -> None:
        super().on_key_released_event(controller, keyval, keycode, state)

        self.check_login_availability()

        if keyval == Gdk.KEY_Return:
            if self.username and self.__password:
                self.login_button.emit('clicked')
            else:
                self.status.error(_("Username or Password is blank!"))

    async def on_login_button_clicked(
            self,
            auto: bool,
            auth_code: str = '',
            auth_code_type: Optional[AuthCodeType] = AuthCodeType.device,
    ) -> None:
        self.status.info(_("Retrieving user data"))
        self.application.main_window.statusbar.set_warning("steamguard", _("Not logged in"))
        self.username_item.set_sensitive(False)
        self.__password_item.set_sensitive(False)
        self.save_password_item.set_sensitive(False)
        self.login_button.set_sensitive(False)

        _login_session = login.Login.get_session(0)
        _login_session.username = self.username
        _login_session.password = self.__password

        if not self.shared_secret:
            log.warning(_("No shared secret found. Trying to log-in without two-factor authentication."))
            # self.code_item.set_visible(True)

        self.status.info(_("Logging in"))
        self.auth_code_item.set_visible(False)

        try_count = 3

        while True:
            try:
                login_data = await _login_session.do_login(
                    self.shared_secret,
                    self.auth_code if self.auth_code else auth_code,
                    auth_code_type,
                    self.mobile_login,
                )
            except login.MailCodeError:
                self.status.info(_("Write code received by email\nand click on 'Log-in' button"))
                self.auth_code_item.set_text("")
                self.auth_code_item.set_visible(True)
                self.auth_code_item.grab_focus()
                auth_code_type = AuthCodeType.email
            except login.TwoFactorCodeError:
                if self.shared_secret:
                    if try_count > 0:
                        for count in range(10, 0):
                            self.status.info(_("Retrying login in {} seconds").format(count))
                            await asyncio.sleep(1)

                        try_count -= 1
                        continue
                    else:
                        self.status.error(_("shared secret is invalid!"))
                        self.username_item.set_sensitive(True)
                        self.__password_item.set_sensitive(True)
                        self.shared_secret_item.grab_focus()
                        break

                self.status.error(_("Write Steam Code bellow and click on 'Log-in'"))
                self.auth_code_item.set_text("")
                self.auth_code_item.set_visible(True)
                self.auth_code_item.grab_focus()
                auth_code_type = AuthCodeType.device
            except login.LoginBlockedError:
                self.status.error(_(
                    "Your network is blocked!\n"
                    "It'll take some time until unblocked. Please, try again later\n"
                ))
                self.user_details_section.set_visible(False)
                self.advanced_login.set_visible(False)
                self.advanced_login_section.set_visible(False)
                self.login_button.set_visible(False)
                self.set_deletable(True)
            except login.CaptchaError as exception:
                # TODO: Captcha gid?? (where did you go? where did you go?)
                self.status.info(_("Write captcha code as shown bellow\nand click on 'Log-in' button"))

                # pixbuf_loader = GdkPixbuf.PixbufLoader()
                # pixbuf_loader.set_size(140, 50)
                # pixbuf_loader.write(exception.captcha)
                # pixbuf_loader.close()
                # self.captcha_item.set_from_pixbuf(pixbuf_loader.get_pixbuf())

                # self.captcha_item.set_visible(True)
                self.captcha_text_item.set_text("")
                self.captcha_text_item.set_visible(True)
                self.captcha_text_item.grab_focus()
            except (login.LoginError, AttributeError) as exception:
                log.error(str(exception))
                # self.__password_item.set_text('')
                self.__password_item.grab_focus()
                config.remove('login', 'refresh_token')
                config.remove('login', 'access_token')

                self.status.error(
                    ':\n'.join(str(exception).split(': ')) +
                    _(
                        '\n\nIf your previous authenticator has been removed,'
                        '\nopen advanced login bellow and remove the old secrets.'
                    ),
                )

                self.username_item.set_sensitive(True)
                self.__password_item.set_sensitive(True)
                self.__password_item.grab_focus()
            except binascii.Error:
                self.status.error(_("shared secret is invalid!"))
                self.username_item.set_sensitive(True)
                self.__password_item.set_sensitive(True)
                self.shared_secret_item.grab_focus()
            except (aiohttp.ClientError, ValueError):
                for count in range(20, 0):
                    self.status.error(_("Check your connection. (server down? blocked?)\nWaiting {}").format(count))
                    await asyncio.sleep(1)

                self.username_item.set_sensitive(True)
                self.__password_item.set_sensitive(True)
                self.login_button.grab_focus()
            else:
                new_configs = {
                    "account_name": self.username,
                    'steamid': login_data.steamid,
                    'refresh_token': login_data.refresh_token,
                    'access_token': login_data.access_token,
                }

                if self.save_password_item.get_active():
                    encrypted_password = core.utils.encode_password(self.__password)
                    new_configs["password"] = encrypted_password

                for key_, value_ in new_configs.items():
                    config.new("login", key_, value_)

                # steamid = universe.generate_steamid(new_configs['steamid'])
                # _login_session.restore_login(steamid, new_configs['token'], new_configs['token_secure'])

                self.application.main_window.statusbar.clear('steamguard')

                self.has_user_data = True
                self.destroy()
            finally:
                self.save_password_item.set_sensitive(True)
                self.login_button.set_sensitive(True)
                self.set_size_request(400, 100)

            break

    def on_advanced_login_clicked(self, *args: Any) -> None:
        if self.advanced_login_section.props.visible:
            self.identity_secret_item.set_text('')
            self.shared_secret_item.set_text('')
            self.advanced_login_section.set_visible(False)
            self.set_size_request(400, 100)
        else:
            self.advanced_login_section.set_visible(True)



#!/usr/bin/env python
#
# Lara Maia <dev@lara.monster> 2015 ~ 2023
#
# The Steam Tools NG is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Steam Tools NG is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.
#
import asyncio
import binascii
import getpass
import logging
from typing import TYPE_CHECKING, Optional

import aiohttp

from stlib import login
from stlib.login import AuthCodeType
from . import utils
from .. import i18n, config, core

if TYPE_CHECKING:
    from . import cli

log = logging.getLogger(__name__)
_ = i18n.get_translation


# noinspection PyUnusedLocal
class Login:
    def __init__(self, cli_: 'cli.SteamToolsNG', mobile_login: bool = True) -> None:
        self.cli = cli_
        self.mobile_login = mobile_login
        self.has_user_data = False
        self._username = ''
        self.__password = ''

    @property
    def username(self) -> str:
        return self._username

    def set_password(self, encrypted_password: str) -> None:
        try:
            __password = core.utils.decode_password(encrypted_password)
            self.__password = __password
        except (binascii.Error, UnicodeError, TypeError):
            log.warning(_("Password decode failed. Trying RAW."))
            self.__password = encrypted_password

    @property
    def shared_secret(self) -> str:
        return config.parser.get("login", "shared_secret")

    @property
    def identity_secret(self) -> str:
        return config.parser.get("login", "identity_secret")

    async def do_login(
            self,
            auto: bool,
            auth_code: str = '',
            auth_code_type: Optional[AuthCodeType] = AuthCodeType.device,
    ) -> None:
        utils.set_console(info=_("Retrieving user data"))

        if auto:
            self._username = config.parser.get("login", "account_name")
            encrypted_password = config.parser.get("login", "password")
            self.set_password(encrypted_password)

        if not self.username or not self.__password:
            user_input = utils.safe_input(_("Please, write your username"))
            assert isinstance(user_input, str), "Safe input is returning bool when it should return str"
            config.new("login", "account_name", user_input)
            self._username = user_input

            self.__password = getpass.getpass(_("Please, write your password (IT'S HIDDEN, and will be encrypted)"))
            encrypted_password = core.utils.encode_password(self.__password)
            config.new("login", "password", encrypted_password)

        _login_session = login.Login.get_session(0)
        _login_session.username = self.username
        _login_session.password = self.__password

        if not self.shared_secret:
            log.warning(_("No shared secret found. Trying to log-in without two-factor authentication."))

        utils.set_console(info=_("Logging in"))
        try_count = 3

        while True:
            try:
                login_data = await _login_session.do_login(
                    self.shared_secret,
                    auth_code,
                    auth_code_type,
                    self.mobile_login,
                )
            except login.MailCodeError:
                user_input = utils.safe_input(_("Write code received by email"))
                assert isinstance(user_input, str), "safe_input is returning bool when it should return str"
                await self.do_login(True, user_input, AuthCodeType.email)
            except login.TwoFactorCodeError:
                if self.shared_secret:
                    if try_count > 0:
                        log.warning(_("Retrying login in 10 seconds"))
                        await asyncio.sleep(10)
                        try_count -= 1
                        continue
                    else:
                        log.error(_("shared secret is invalid!"))
                        self.cli.on_quit()

                user_input = utils.safe_input(_("Write Steam Code"))
                assert isinstance(user_input, str), "safe_input is returning bool when it should return str"
                await self.do_login(True, user_input)
            except login.LoginBlockedError:
                log.error(_(
                    "Your network is blocked!\n"
                    "It'll take some time until unblocked. Please, try again later\n"
                ))
                self.cli.on_quit()
            except login.CaptchaError as exception:
                utils.set_console(info=_("Steam server is requesting a captcha code."))
                # TODO: Captcha gid?? (where did you go? where did you go?)
                # with tempfile.NamedTemporaryFile(buffering=0, prefix='stng_', suffix='.captcha.png') as temp_file:
                #    temp_file.write(exception.captcha)
                #    temp_file.flush()
                user_input = utils.safe_input(_("Write Captcha Code"))
                assert isinstance(user_input, str)
                await self.do_login(True, user_input, AuthCodeType.machine)
            except binascii.Error:
                log.error(_("shared secret is invalid!"))
                self.cli.on_quit()
            except login.LoginError as exception:
                log.error(str(exception))
                config.remove('login', 'refresh_token')
                config.remove('login', 'access_token')
                log.warning(_(
                    "If your previous authenticator has been removed,"
                    "\nopen your config file and remove the old secrets."
                ))
                self.cli.on_quit()
            except (aiohttp.ClientError, ValueError):
                log.error(_("Check your connection. (server down?)"))
                await asyncio.sleep(15)
                continue
            else:
                new_configs = {
                    "account_name": self.username,
                    'steamid': login_data.steamid,
                    'refresh_token': login_data.refresh_token,
                    'access_token': login_data.access_token,
                }

                for key, value in new_configs.items():
                    config.new("login", key, value)

                # steamid = universe.generate_steamid(new_configs['steamid'])
                # _login_session.restore_login(steamid, new_configs['token'], new_configs['token_secure'])
                self.has_user_data = True

            break



#!/usr/bin/env python
#
# Lara Maia <dev@lara.monster> 2015 ~ 2023
#
# The Steam Tools NG is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Steam Tools NG is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.
#
import asyncio
import contextlib
import random
import time
from subprocess import call
from typing import AsyncGenerator, Dict, Optional, Any

import aiohttp

from stlib import webapi, client, universe, community
from . import utils
from .. import i18n, config

_ = i18n.get_translation
executors = {}


def safe_exit(*args: Any, **kwargs: Any) -> None:
    for executor in executors.values():
        with contextlib.suppress(RuntimeError):
            executor.shutdown(*args, **kwargs)


async def while_has_cards(
        steamid: universe.SteamId,
        badge: community.Badge,
        play_event: Optional[asyncio.Event] = None,
) -> AsyncGenerator[utils.ModuleData, None]:
    webapi_session = webapi.SteamWebAPI.get_session(0)
    community_session = community.Community.get_session(0)

    while badge.cards != 0:
        if play_event:
            await play_event.wait()

        mandatory_waiting = config.parser.getint("cardfarming", "mandatory_waiting")
        wait_while_running = config.parser.getint("cardfarming", "wait_while_running")
        wait_for_drops = config.parser.getint("cardfarming", "wait_for_drops")

        try:
            game_list = await webapi_session.get_owned_games(steamid, appids_filter=[badge.appid])
            game_info = game_list[0]
        except aiohttp.ClientError:
            module_data = utils.ModuleData(error=_("Check your connection. (server down?)"), info=_("Waiting Changes"))

            async for data in utils.timed_module_data(10, module_data):
                yield data

            continue

        if game_info.playtime_forever * 60 >= mandatory_waiting:
            wait_offset = random.randint(wait_while_running, int(wait_while_running / 100 * 125))
        else:
            wait_offset = mandatory_waiting - game_info.playtime_forever * 60

        try:
            executor = client.SteamAPIExecutor(badge.appid)
        except AttributeError:
            yield utils.ModuleData(action='ignore', info=_("Invalid game id {}. Ignoring.").format(badge.appid))
            break
        except ProcessLookupError:
            module_data = utils.ModuleData(error=_("Steam Client is not running."), info=_("Waiting Changes"))

            async for data in utils.timed_module_data(15, module_data):
                yield data

            continue

        module_data = utils.ModuleData(
            display=str(badge.appid),
            info=badge.name,
            status=_("Running {}").format(badge.name),
            raw_data=executor,
            action="check",
        )

        async for data in utils.timed_module_data(wait_offset, module_data):
            if play_event and not play_event.is_set():
                executor.shutdown()
                await asyncio.sleep(1)
                await play_event.wait()
                executor.__init__(executor.appid)

            yield data

        executor.shutdown()
        wait_offset = random.randint(wait_for_drops, int(wait_for_drops / 100 * 125))

        module_data = utils.ModuleData(
            display=str(badge.appid),
            info=_('Updating {} drops').format(badge.name),
            status=_("Game paused"),
        )

        async for data in utils.timed_module_data(wait_offset, module_data):
            yield data

        while True:
            try:
                cards = await community_session.get_card_drops_remaining(steamid, badge.appid)
            except aiohttp.ClientError:
                yield utils.ModuleData(error=_("Check your connection. (server down?)"), info=_("Waiting Changes"))
                await asyncio.sleep(10)
            except community.BadgeError:
                yield utils.ModuleData(error=_("Steam Server is busy"), info=_("Waiting Changes"))
                await asyncio.sleep(20)
            else:
                break

        yield utils.ModuleData(action="update_drops", raw_data=badge.cards - cards)

        # noinspection PyProtectedMember
        badge = badge._replace(cards=cards)

    utils.ModuleData(
        display=str(badge.appid),
        info=_("{} ({})").format(_("Done"), badge.name),
    )


async def main(
        steamid: universe.SteamId,
        play_event: Optional[asyncio.Event] = None,
        custom_game_id: int = 0,
) -> AsyncGenerator[utils.ModuleData, None]:
    if play_event:
        await play_event.wait()

    asyncio.current_task().add_done_callback(safe_exit)

    reverse_sorting = config.parser.getboolean("cardfarming", "reverse_sorting")
    max_concurrency = config.parser.getint("cardfarming", "max_concurrency")
    invisible = config.parser.getboolean("cardfarming", "invisible")
    community_session = community.Community.get_session(0)
    total_cards_remaining = 0

    try:
        badges = sorted(
            await community_session.get_badges(steamid),
            key=lambda badge_: badge_.cards,  # type: ignore
            reverse=reverse_sorting
        )
    except aiohttp.ClientError:
        module_data = utils.ModuleData(error=_("Check your connection. (server down?)"), info=_("Waiting Changes"))

        async for data in utils.timed_module_data(10, module_data):
            yield data

        return

    if not badges or (custom_game_id and custom_game_id not in [badge.appid for badge in badges]):
        module_data = utils.ModuleData(error=_("No more cards to drop."), info=_("Waiting Changes"))
        wait_offset = random.randint(300, 500)

        async for data in utils.timed_module_data(wait_offset, module_data):
            yield data

        return

    generators = {}

    if invisible:
        call(f'{config.file_manager} "steam://friends/status/invisible"')

    for badge in badges:
        yield utils.ModuleData(
            display=str(badge.appid),
            status=_("Loading {}").format(badge.name),
        )

        if custom_game_id and badge.appid != custom_game_id:
            yield utils.ModuleData(info=_("Skipping {}").format(badge.appid))
            continue

        generators[badge.appid] = while_has_cards(steamid, badge, play_event)
        total_cards_remaining += badge.cards

    tasks: Dict[int, Optional[asyncio.Task[Any]]] = {}
    semaphore = asyncio.Semaphore(max_concurrency)
    last_update = 0

    while True:
        for appid in generators:
            progress_coro = anext(generators[appid])
            assert asyncio.iscoroutine(progress_coro)

            if appid not in tasks:
                if semaphore.locked():
                    break

                await semaphore.acquire()
                tasks[appid] = asyncio.create_task(progress_coro)

            if not tasks[appid]:
                continue

            current_task = tasks[appid]
            assert isinstance(current_task, asyncio.Task)

            if current_task.done():
                semaphore.release()

                if current_task.exception():
                    if isinstance(current_task.exception(), StopAsyncIteration):
                        tasks[appid] = None
                        continue

                    current_exception = current_task.exception()
                    assert isinstance(current_exception, BaseException)
                    raise current_exception

                await semaphore.acquire()
                tasks[appid] = asyncio.create_task(progress_coro)

        if not any(tasks.values()):
            break

        await asyncio.wait([task for task in tasks.values() if task], return_when=asyncio.FIRST_COMPLETED)

        for appid, task in tasks.items():
            if task and task.done() and not task.exception():
                global executors
                data: utils.ModuleData = task.result()

                if data.action == 'check':
                    executors[appid] = data.raw_data

                if data.action == "update_drops":
                    total_cards_remaining -= data.raw_data

                if int(time.time()) > last_update + 3:
                    current_running_limit = len(tasks)
                    total_remaining = len(generators) - len([task for task in tasks.values() if not task])
                    running_executors = [executor for executor in executors.values() if executor.is_running()]
                    extra_info = ''

                    current_running_limit = min(current_running_limit, total_remaining)
                    if current_running_limit == 2:
                        extra_info = _(" +{} other").format(current_running_limit - 1)
                    elif current_running_limit > 2:
                        extra_info = _(" +{} others").format(current_running_limit - 1)

                    yield utils.ModuleData(
                        display=' : '.join([str(executor.appid) for executor in running_executors]),
                        info=data.info + extra_info,
                        status=_('{} from {} remaining ({} cards)').format(
                            current_running_limit,
                            total_remaining,
                            total_cards_remaining,
                        ),
                        level=data.level,
                        raw_data=running_executors,
                        action=data.action,
                    )
                    last_update = int(time.time())



#!/usr/bin/env python
#
# Lara Maia <dev@lara.monster> 2015 ~ 2023
#
# The Steam Tools NG is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Steam Tools NG is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.
#
import asyncio
import logging
from typing import Any, Dict, List, Tuple

from gi.repository import Gtk

from stlib import universe, community
from . import utils
from .. import config, i18n

log = logging.getLogger(__name__)
_ = i18n.get_translation


# noinspection PyUnusedLocal
class FinalizeWindow(utils.PopupWindowBase):
    def __init__(
            self,
            parent_window: Gtk.Window,
            application: Gtk.Application,
            confirmations_tree: utils.SimpleTextTree,
            action: str,
            batch: bool = False,
    ) -> None:
        super().__init__(parent_window, application)
        self.community_session = community.Community.get_session(0)
        self.confirmations_tree = confirmations_tree
        self.selection = self.confirmations_tree.model.get_selected_item()
        self.batch = batch

        self.action = _("accept") if action == "allow" else _("cancel")
        self.raw_action = action
        self.header_bar.set_show_title_buttons(False)
        self.set_title(_('Finalize Confirmation'))

        self.status = utils.SimpleStatus()
        self.content_grid.attach(self.status, 0, 0, 1, 1)

        self.progress = Gtk.LevelBar()
        self.progress.set_visible(False)
        self.content_grid.attach(self.progress, 0, 1, 1, 1)

        self.yes_button = Gtk.Button()
        self.yes_button.set_label(_("Continue"))
        self.yes_button.connect("clicked", self.on_yes_button_clicked)
        self.header_bar.pack_end(self.yes_button)

        self.no_button = Gtk.Button()
        self.no_button.set_label(_("Cancel"))
        self.no_button.connect("clicked", lambda button: self.destroy())
        self.header_bar.pack_start(self.no_button)

        if self.batch:
            self.status.info(
                _("Do you really want to {} ALL confirmations?\nIt can't be undone!").format(self.action.upper())
            )
        elif self.selection:
            self.status.info(
                _("{}\nDo you want to {} the offer?\nIt can't be undone!").format(
                    self.selection.get_item().summary,
                    self.action.upper(),
                    utils.unmarkup(self.selection.get_item().to),
                )
            )
        else:
            self.status.error(_("You must select something"))
            self.header_bar.set_show_title_buttons(True)
            self.yes_button.set_visible(False)
            self.no_button.set_visible(False)

    def on_yes_button_clicked(self, button: Gtk.Button) -> None:
        button.set_visible(False)
        self.no_button.set_visible(False)
        self.progress.set_visible(True)
        self.set_size_request(0, 0)
        self.header_bar.set_show_title_buttons(False)
        self.confirmations_tree.lock = True

        loop = asyncio.get_event_loop()

        if self.batch:
            task = loop.create_task(self.batch_finalize())
        else:
            task = loop.create_task(self.single_finalize())

        task.add_done_callback(self.on_task_finish)

    def on_task_finish(self, task: asyncio.Task[Any]) -> None:
        self.progress.set_visible(False)
        self.confirmations_tree.lock = False
        exception = task.exception()

        if exception and not isinstance(exception, asyncio.CancelledError):
            try:
                current_exception = task.exception()
                assert isinstance(current_exception, BaseException)
                raise current_exception
            except Exception as exception:
                stack = task.get_stack()

                for frame in stack:
                    log.error("%s at %s", type(exception).__name__, frame)

                log.error("Steam Server is slow. (%s)", str(exception))

                self.status.error(
                    _(
                        "Unable to complete this confirmation. The reason is one of the following:\n\n"
                        "1. The confirmation you choose already gone. Try another one.\n"
                        "2. You wrote a wrong token in config. Update you config.\n"
                        "3. The Steam server is slow. Wait a minute and try again.\n\n"
                        "If you keep seeing this error, please update the confirmation list."
                    )
                )

                self.header_bar.set_show_title_buttons(True)
                self.yes_button.set_visible(False)
        else:
            self.destroy()

    async def do_finalize(self, item: Gtk.ListItem) -> Dict[str, Any]:
        identity_secret = config.parser.get("login", "identity_secret")
        deviceid = config.parser.get("login", "deviceid")
        steamid_raw = config.parser.getint("login", "steamid")

        try:
            steamid = universe.generate_steamid(steamid_raw)
        except ValueError:
            self.status.info(_("Your steam is invalid. (are you logged in?)"))
            await asyncio.sleep(5)
            return {}

        self.status.info(_("Waiting Steam Server (OP: {})").format(item.creatorid))
        result: Dict[str, Any] = {}

        # steam confirmation server isn't reliable
        for i in range(2):
            result = await self.community_session.send_confirmation(
                identity_secret,
                steamid,
                deviceid,
                item.id,
                item.nonce,
                self.raw_action,
            )
            await asyncio.sleep(0.5)

        assert isinstance(result, dict)
        return result

    async def single_finalize(self) -> Dict[str, Any]:
        item = self.selection.get_item()
        result = await self.do_finalize(item)
        self.confirmations_tree.remove_row(self.selection)

        assert isinstance(result, dict)
        return result

    async def batch_finalize(self) -> List[Tuple[Gtk.TreeIter, Dict[str, Any]]]:
        results = []
        n_items = self.confirmations_tree.store.get_n_items()
        self.status.info(_("Waiting Steam Server response"))

        for index in range(n_items):
            self.progress.set_value(index)
            self.progress.set_max_value(n_items)
            item = self.confirmations_tree.store.get_item(index)
            result = await self.do_finalize(item)
            results.append((item, result))

        self.confirmations_tree.clear()

        return results



#!/usr/bin/env python
#
# Lara Maia <dev@lara.monster> 2015 ~ 2023
#
# The Steam Tools NG is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Steam Tools NG is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.
#

import asyncio
import contextlib
import logging
from subprocess import call
from typing import Optional, Tuple, Any

from gi.repository import Gio, Gtk, Gdk

import stlib
from stlib import login, plugins
from . import confirmation, utils, coupon, authenticator
from .login import LoginWindow
from .. import config, i18n, core

_ = i18n.get_translation
log = logging.getLogger(__name__)

if stlib.steamworks_available:
    from stlib import client


# noinspection PyUnusedLocal
class Main(Gtk.ApplicationWindow):
    def __init__(self, application: Gtk.Application, title: str) -> None:
        super().__init__(application=application, title=title)
        self.application = application

        _display = Gdk.Display.get_default()
        _style_provider = Gtk.CssProvider()

        _style_provider.load_from_data(
            b"* { border-radius: 2px; }"
            b"label.warning { background-color: darkblue; color: white; }"
            b"label.critical { background-color: darkred; color: white; }"
        )

        _style_context = self.get_style_context()

        _style_context.add_provider_for_display(
            _display,
            _style_provider,
            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION,
        )

        header_bar = Gtk.HeaderBar()

        menu = Gio.Menu()
        menu.append(_("Settings"), "app.settings")
        menu.append(_("About"), "app.about")
        menu.append(_("Exit"), "app.exit")

        menu_button = Gtk.MenuButton()
        menu_button.set_icon_name("open-menu")
        menu_button.set_menu_model(menu)
        header_bar.pack_end(menu_button)

        self.set_default_size(750, 10)
        self.set_resizable(False)

        if config.parser.getboolean("general", "show_close_button"):
            self.set_deletable(True)
        else:
            self.set_deletable(False)

        self.set_titlebar(header_bar)
        self.set_title('Steam Tools NG')

        main_grid = Gtk.Grid()
        main_grid.set_margin_start(10)
        main_grid.set_margin_end(10)
        main_grid.set_margin_top(10)
        main_grid.set_margin_bottom(10)
        self.set_child(main_grid)

        self.limited_label = Gtk.Label()
        self.limited_label.set_visible(False)
        self.limited_label.set_markup(
            utils.markup(
                _("Limited Account! Some modules will not work!"),
                background='red',
                color='white',
            )
        )
        main_grid.attach(self.limited_label, 1, 0, 1, 1)

        self.user_info_label = Gtk.Label()
        self.user_info_label.set_halign(Gtk.Align.END)
        header_bar.pack_start(self.user_info_label)

        self.main_tabs = Gtk.Stack()
        self.main_tabs.set_hhomogeneous(True)
        self.main_tabs.connect("notify::visible-child", self.on_stack_child_changed)
        main_grid.attach(self.main_tabs, 1, 2, 1, 1)

        switcher = Gtk.StackSwitcher()
        switcher.set_stack(self.main_tabs)
        main_grid.attach(switcher, 1, 1, 1, 1)

        steamguard_section = utils.Section("steamguard")
        steamguard_section.stackup_section("SteamGuard", self.main_tabs)

        cardfarming_section = utils.Section("cardfarming")
        cardfarming_section.stackup_section("CardFarming", self.main_tabs)

        steamgifts_section = utils.Section("steamgifts")
        steamgifts_section.stackup_section("SteamGifts", self.main_tabs)

        steamtrades_section = utils.Section("steamtrades")
        steamtrades_section.stackup_section("SteamTrades", self.main_tabs)

        coupons_section = utils.Section("coupons")
        coupons_section.stackup_section(_("Coupons"), self.main_tabs)

        # grid managed by plugin switch
        self.steamguard_status = utils.Status(4)
        self.cardfarming_status = utils.Status(6)
        self.steamgifts_status = utils.Status(5)
        self.steamtrades_status = utils.Status(5)

        steamguard_section.grid.attach(self.steamguard_status, 0, 0, 2, 1)

        steamguard_stack = Gtk.Stack()
        steamguard_stack.set_vexpand(True)
        steamguard_section.grid.attach(steamguard_stack, 1, 1, 1, 1)

        steamguard_sidebar = Gtk.StackSidebar()
        steamguard_sidebar.set_stack(steamguard_stack)
        steamguard_sidebar.set_size_request(150, -1)
        steamguard_section.grid.attach(steamguard_sidebar, 0, 1, 1, 1)

        self.confirmations_grid = Gtk.Grid()
        self.confirmations_grid.set_row_spacing(10)
        steamguard_stack.add_titled(self.confirmations_grid, "confirmations", _("Confirmations"))

        confirmation_tree_headers = 'id', 'creatorid', 'nonce', '_give', '_to', '_receive', 'summary',
        self.confirmations_tree = utils.SimpleTextTree(*confirmation_tree_headers)
        self.confirmations_grid.attach(self.confirmations_tree, 0, 0, 4, 1)

        for index, column in enumerate(self.confirmations_tree.view.get_columns()):
            column.set_resizable(True)

            if index in (1, 2, 3, 7):
                column.set_visible(False)

            if index in (4, 6):
                column.set_fixed_width(190)

            if index == 5:
                column.set_fixed_width(100)

        self.confirmations_tree.model.connect("selection-changed", self.on_tree_selection_changed)

        accept_button = Gtk.Button()
        accept_button.set_margin_start(3)
        accept_button.set_margin_end(3)
        accept_button.set_label(_('Accept'))
        accept_button.connect('clicked', self.on_validate_confirmations, "allow")
        self.confirmations_grid.attach(accept_button, 0, 1, 1, 1)

        cancel_button = Gtk.Button()
        cancel_button.set_margin_start(3)
        cancel_button.set_margin_end(3)
        cancel_button.set_label(_('Cancel'))
        cancel_button.connect('clicked', self.on_validate_confirmations, "cancel")
        self.confirmations_grid.attach(cancel_button, 1, 1, 1, 1)

        accept_all_button = Gtk.Button()
        accept_all_button.set_margin_start(3)
        accept_all_button.set_margin_end(3)
        accept_all_button.set_label(_('Accept All'))
        accept_all_button.connect('clicked', self.on_validate_confirmations, "allow", True)
        self.confirmations_grid.attach(accept_all_button, 2, 1, 1, 1)

        cancel_all_button = Gtk.Button()
        cancel_all_button.set_margin_start(3)
        cancel_all_button.set_margin_end(3)
        cancel_all_button.set_label(_('Cancel All'))
        cancel_all_button.connect('clicked', self.on_validate_confirmations, "cancel", True)
        self.confirmations_grid.attach(cancel_all_button, 3, 1, 1, 1)

        steamguard_settings = utils.Section("steamguard")
        steamguard_settings.stackup_section(_("Settings"), steamguard_stack)
        steamguard_settings.grid.set_halign(Gtk.Align.CENTER)

        self.steamguard_enable = steamguard_settings.new_item("enable", _("Enable:"), Gtk.Switch, 0, 0)
        self.steamguard_enable.set_margin_top(40)
        self.steamguard_enable.label.set_margin_top(40)
        self.steamguard_enable.set_halign(Gtk.Align.END)
        self.steamguard_enable.connect("state-set", utils.on_setting_state_set)

        self.steamguard_disabled = Gtk.Label()
        self.steamguard_disabled.set_justify(Gtk.Justification.CENTER)
        self.steamguard_disabled.set_halign(Gtk.Align.CENTER)

        _message = _(
            "steamguard module has been disabled because you have\n"
            "logged in but no shared secret is found. To enable it again,\n"
            "go to Advanced and add a valid shared secret\n"
            "or use STNG as your Steam Authenticator\n"
        )

        self.steamguard_disabled.set_markup(utils.markup(_message, color="hotpink", background="black"))
        self.steamguard_disabled.set_visible(False)
        steamguard_section.grid.attach(self.steamguard_disabled, 0, 0, 2, 1)

        self.confirmations_enable = steamguard_settings.new_item(
            "enable_confirmations", _("Enable Confirmations:"),
            Gtk.Switch,
            0, 1,
        )

        self.confirmations_enable.set_halign(Gtk.Align.END)
        self.confirmations_enable.connect("state-set", utils.on_setting_state_set)

        self.confirmations_disabled = Gtk.Label()
        self.confirmations_disabled.set_justify(Gtk.Justification.CENTER)
        self.confirmations_disabled.set_halign(Gtk.Align.CENTER)

        _message = _(
            "confirmations module has been disabled because you have\n"
            "logged in but no identity secret is found. To enable it again,\n"
            "go to login -> advanced and add a valid identity secret\n"
            "or use STNG as your Steam Authenticator\n"
        )

        self.confirmations_disabled.set_markup(utils.markup(_message, color="hotpink", background="black"))
        self.confirmations_disabled.set_visible(False)
        self.confirmations_grid.attach(self.confirmations_disabled, 0, 0, 4, 1)

        login_button = Gtk.Button()
        login_button.set_margin_top(40)
        login_button.set_label(_("Login with another account"))
        login_button.set_name("login_button")
        login_button.connect('clicked', self.on_login_button_clicked)
        steamguard_settings.grid.attach(login_button, 0, 2, 2, 1)

        new_authenticator_button = Gtk.Button()
        new_authenticator_button.set_label(_("Add STNG as your Steam Authenticator"))
        new_authenticator_button.set_name("new_authenticator_button")
        new_authenticator_button.connect("clicked", self.on_new_authenticator_clicked)
        steamguard_settings.grid.attach(new_authenticator_button, 0, 3, 2, 1)

        remove_authenticator_button = Gtk.Button()
        remove_authenticator_button.set_label(_("Remove STNG Authenticator from your Steam Account"))
        remove_authenticator_button.set_name("remove_authenticator_button")
        remove_authenticator_button.connect("clicked", self.on_remove_authenticator_clicked)
        steamguard_settings.grid.attach(remove_authenticator_button, 0, 4, 2, 1)

        reset_password_button = Gtk.Button()
        reset_password_button.set_label(_("Remove Saved Password"))
        reset_password_button.set_name("reset_password_button")
        reset_password_button.connect("clicked", self.on_reset_password_clicked)
        steamguard_settings.grid.attach(reset_password_button, 0, 5, 2, 1)

        steamguard_advanced = utils.Section("login")
        steamguard_advanced.stackup_section(_("Advanced"), steamguard_stack, scroll=True)

        warning_label = Gtk.Label()
        warning_label.set_markup(utils.markup(
            _("Warning: Don't mess up these settings unless you know what you are doing!"),
            color='darkred' if self.theme == 'light' else 'red',
        ))
        steamguard_advanced.grid.attach(warning_label, 0, 0, 2, 1)

        shared_secret = steamguard_advanced.new_item('shared_secret', _("Shared Secret:"), Gtk.Entry, 0, 1)
        shared_secret.connect('changed', utils.on_setting_changed)

        access_token_item = steamguard_advanced.new_item("access_token", _("Access Token:"), Gtk.Entry, 0, 2)
        access_token_item.connect("changed", utils.on_setting_changed)

        refresh_token_item = steamguard_advanced.new_item("refresh_token", _("Refresh Token:"), Gtk.Entry, 0, 3)
        refresh_token_item.connect("changed", utils.on_setting_changed)

        identity_secret = steamguard_advanced.new_item('identity_secret', _("Identity Secret:"), Gtk.Entry, 0, 4)
        identity_secret.connect('changed', utils.on_setting_changed)

        deviceid = steamguard_advanced.new_item('deviceid', _("Device ID:"), Gtk.Entry, 0, 5)
        deviceid.connect('changed', utils.on_setting_changed)

        steamid_item = steamguard_advanced.new_item("steamid", _("Steam ID:"), Gtk.Entry, 0, 6)
        steamid_item.set_input_purpose(Gtk.InputPurpose.DIGITS)
        steamid_item.connect("changed", utils.on_digit_only_setting_changed)

        account_name = steamguard_advanced.new_item('account_name', _("Username:"), Gtk.Entry, 0, 7)
        account_name.connect('changed', utils.on_setting_changed)

        reset_button = Gtk.Button()
        reset_button.set_label(_("Reset Everything (USE WITH CAUTION!!!)"))
        reset_button.set_name("reset_button")
        reset_button.connect("clicked", self.on_reset_clicked)
        steamguard_advanced.grid.attach(reset_button, 0, 8, 2, 1)

        cardfarming_section.grid.attach(self.cardfarming_status, 0, 0, 2, 1)

        cardfarming_stack = Gtk.Stack()
        cardfarming_stack.set_vexpand(True)
        cardfarming_section.grid.attach(cardfarming_stack, 1, 1, 1, 1)

        cardfarming_sidebar = Gtk.StackSidebar()
        cardfarming_sidebar.set_stack(cardfarming_stack)
        cardfarming_sidebar.set_size_request(150, -1)
        cardfarming_section.grid.attach(cardfarming_sidebar, 0, 1, 1, 1)

        cardfarming_settings = utils.Section("cardfarming")
        cardfarming_settings.stackup_section(_("Settings"), cardfarming_stack, scroll=True)

        cardfarming_enable = cardfarming_settings.new_item("enable", _("Enable:"), Gtk.Switch, 0, 1)
        cardfarming_enable.set_halign(Gtk.Align.END)
        cardfarming_enable.connect("state-set", utils.on_setting_state_set)

        mandatory_waiting = cardfarming_settings.new_item("mandatory_waiting", _("Mandatory waiting:"), Gtk.Entry, 0, 2)
        mandatory_waiting.connect("changed", utils.on_digit_only_setting_changed)

        wait_while_running = cardfarming_settings.new_item(
            "wait_while_running",
            _("Wait while running:"),
            Gtk.Entry,
            0, 3,
        )
        wait_while_running.connect("changed", utils.on_digit_only_setting_changed)

        wait_for_drops = cardfarming_settings.new_item("wait_for_drops", _("Wait for drops:"), Gtk.Entry, 0, 4)
        wait_for_drops.connect("changed", utils.on_digit_only_setting_changed)

        max_concurrency = cardfarming_settings.new_item("max_concurrency", _("Max concurrency:"), Gtk.Entry, 0, 5)
        max_concurrency.connect("changed", utils.on_digit_only_setting_changed)

        cardfarming_invisible = cardfarming_settings.new_item("invisible", _("Invisible:"), Gtk.Switch, 0, 6)
        cardfarming_invisible.set_halign(Gtk.Align.END)
        cardfarming_invisible.connect("state-set", utils.on_setting_state_set)

        reverse_sorting = cardfarming_settings.new_item("reverse_sorting", _("Reverse Sorting:"), Gtk.Switch, 0, 7)
        reverse_sorting.set_halign(Gtk.Align.END)
        reverse_sorting.connect("state-set", utils.on_setting_state_set)

        if not stlib.steamworks_available:
            cardfarming_settings.set_sensitive(False)
            cardfarming_enable.set_active(False)
            _cardfarming_disabled = Gtk.Label()
            _cardfarming_disabled.set_justify(Gtk.Justification.CENTER)
            _cardfarming_disabled.set_halign(Gtk.Align.CENTER)

            _message = _(
                "cardfarming module has been disabled because you have\n"
                "a stlib built without SteamWorks support. To enable it again,\n"
                "reinstall stlib with SteamWorks support\n"
            )

            _cardfarming_disabled.set_markup(utils.markup(_message, color="hotpink", background="black"))
            cardfarming_section.attach(_cardfarming_disabled, 0, 0, 2, 1)

        steamgifts_section.grid.attach(self.steamgifts_status, 0, 0, 2, 1)

        steamgifts_stack = Gtk.Stack()
        steamgifts_stack.set_vexpand(True)
        steamgifts_section.grid.attach(steamgifts_stack, 1, 1, 1, 1)

        steamgifts_sidebar = Gtk.StackSidebar()
        steamgifts_sidebar.set_stack(steamgifts_stack)
        steamgifts_sidebar.set_size_request(150, -1)
        steamgifts_section.grid.attach(steamgifts_sidebar, 0, 1, 1, 1)

        steamgifts_settings = utils.Section("steamgifts")
        steamgifts_settings.stackup_section(_("Settings"), steamgifts_stack)

        steamgifts_enable = steamgifts_settings.new_item("enable", _("Enable:"), Gtk.Switch, 0, 0)
        steamgifts_enable.set_halign(Gtk.Align.END)
        steamgifts_enable.connect('state-set', utils.on_setting_state_set)

        developer_giveaways = steamgifts_settings.new_item(
            "developer_giveaways", _("Developer Giveaways"),
            Gtk.Switch,
            0, 1,
        )
        developer_giveaways.set_halign(Gtk.Align.END)
        developer_giveaways.connect("state-set", utils.on_setting_state_set)

        steamgifts_mode = steamgifts_settings.new_item(
            "mode", _("Mode:"),
            Gtk.DropDown,
            0, 2,
            items=config.steamgifts_modes,
        )
        steamgifts_mode.connect("notify::selected", utils.on_dropdown_setting_changed, config.steamgifts_modes)

        wait_after_each_strategy = steamgifts_settings.new_item(
            "wait_after_each_strategy", _("Wait after each strategy:"),
            Gtk.Entry,
            0, 3,
        )
        wait_after_each_strategy.connect("changed", utils.on_digit_only_setting_changed)

        wait_after_full_cycle = steamgifts_settings.new_item(
            "wait_after_full_cycle", _("Wait after full cycle:"),
            Gtk.Entry,
            0, 4,
        )
        wait_after_full_cycle.connect("changed", utils.on_digit_only_setting_changed)

        minimum_points = steamgifts_settings.new_item("minimum_points", _("Minimum points:"), Gtk.Entry, 0, 5)
        minimum_points.connect("changed", utils.on_digit_only_setting_changed)

        for index in range(1, 6):
            strategy_section = utils.Section(f"steamgifts_strategy{index}")
            strategy_section.stackup_section(_("Strategy {}").format(index), steamgifts_stack, scroll=True)

            label = Gtk.Label()
            label.set_text(_("Strategy {}").format(index))
            strategy_section.grid.attach(label, 0, 0, 1, 1)

            enable = strategy_section.new_item("enable", None, Gtk.Switch, 2, 0)
            enable.set_halign(Gtk.Align.END)
            enable.connect("state-set", utils.on_setting_state_set)

            minimum_label = Gtk.Label()
            minimum_label.set_text(_("Minimum"))
            strategy_section.grid.attach(minimum_label, 1, 1, 1, 1)

            maximum_label = Gtk.Label()
            maximum_label.set_text(_("Maximum"))
            strategy_section.grid.attach(maximum_label, 2, 1, 1, 1)

            for tree_level, item in enumerate(["points", "level", "copies", "metascore", "entries"]):
                label = Gtk.Label()
                label.set_text(_(item))
                strategy_section.grid.attach(label, 0, tree_level + 2, 1, 1)

                minimum = strategy_section.new_item(f"minimum_{item}", None, Gtk.Entry, 1, tree_level + 2)
                minimum.connect("changed", utils.on_digit_only_setting_changed)

                maximum = strategy_section.new_item(f"maximum_{item}", None, Gtk.Entry, 2, tree_level + 2)
                maximum.connect("changed", utils.on_digit_only_setting_changed)

            restrict_type = strategy_section.new_item(
                "restrict_type",
                _("Restrict Type:"),
                Gtk.DropDown,
                0, 7,
                items=config.giveaway_types,
            )
            restrict_type.connect("notify::selected", utils.on_dropdown_setting_changed, config.giveaway_types)

            sort_type = strategy_section.new_item(
                "sort_type",
                _("Sort Type:"),
                Gtk.DropDown,
                0, 8,
                items=config.giveaway_sort_types,
            )
            sort_type.connect("notify::selected", utils.on_dropdown_setting_changed, config.giveaway_sort_types)

            # setattr(self, f"steamgifts_strategy{index}", strategy_section)

        if not plugins.has_plugin("steamgifts"):
            steamgifts_section.set_sensitive(False)
            steamgifts_enable.set_active(False)
            _steamgifts_disabled = Gtk.Label()
            _steamgifts_disabled.set_justify(Gtk.Justification.CENTER)
            _steamgifts_disabled.set_halign(Gtk.Align.CENTER)

            _message = _(
                "steamgifts module has been disabled because you don't\n"
                "have steamgifts plugin installed. To enable it again,\n"
                "install the steamgifts plugin.\n"
            )

            _steamgifts_disabled.set_markup(utils.markup(_message, color="hotpink", background="black"))
            steamgifts_section.grid.attach(_steamgifts_disabled, 0, 0, 2, 2)

        steamtrades_section.grid.attach(self.steamtrades_status, 0, 0, 2, 1)

        steamtrades_stack = Gtk.Stack()
        steamtrades_stack.set_vexpand(True)
        steamtrades_section.grid.attach(steamtrades_stack, 1, 1, 1, 1)

        steamtrades_sidebar = Gtk.StackSidebar()
        steamtrades_sidebar.set_stack(steamtrades_stack)
        steamtrades_sidebar.set_size_request(150, -1)
        steamtrades_section.grid.attach(steamtrades_sidebar, 0, 1, 1, 1)

        steamtrades_settings = utils.Section("steamtrades")
        steamtrades_settings.stackup_section(_("Settings"), steamtrades_stack)

        steamtrades_enable = steamtrades_settings.new_item("enable", _("Enable:"), Gtk.Switch, 0, 1)
        steamtrades_enable.label.set_margin_top(40)
        steamtrades_enable.set_margin_top(40)
        steamtrades_enable.set_halign(Gtk.Align.END)
        steamtrades_enable.connect("state-set", utils.on_setting_state_set)

        trade_ids = steamtrades_settings.new_item("trade_ids", _("Trade IDs:"), Gtk.Entry, 0, 2)
        trade_ids.set_placeholder_text('12345, asdfg, ...')
        trade_ids.connect("changed", utils.on_setting_changed)

        wait_for_bump = steamtrades_settings.new_item("wait_for_bump", _("Wait for Bump:"), Gtk.Entry, 0, 3)
        wait_for_bump.connect("changed", utils.on_digit_only_setting_changed)

        if not plugins.has_plugin("steamtrades"):
            steamtrades_settings.set_sensitive(False)
            steamtrades_enable.set_active(False)
            _steamtrades_disabled = Gtk.Label()
            _steamtrades_disabled.set_justify(Gtk.Justification.CENTER)
            _steamtrades_disabled.set_halign(Gtk.Align.CENTER)

            _message = _(
                "steamtrades module has been disabled because you don't\n"
                "have steamtrades plugin installed. To enable it again,\n"
                "install the steamtrades plugin.\n"
            )

            _steamtrades_disabled.set_markup(utils.markup(_message, color="hotpink", background="black"))
            steamtrades_settings.grid.attach(_steamtrades_disabled, 0, 1, 2, 2)

        self.coupon_warning = Gtk.Label()
        self.coupon_warning.set_markup(utils.markup(
            _("Warning: It's a heavy uncached operation. Fetch only once a day or you will be blocked."),
            color='darkred' if self.theme == 'light' else 'red',
        ))
        self.coupon_warning.set_margin_top(37)
        self.coupon_warning.set_margin_bottom(37)
        coupons_section.grid.attach(self.coupon_warning, 0, 0, 2, 1)

        coupons_stack = Gtk.Stack()
        coupons_stack.set_vexpand(True)
        coupons_section.grid.attach(coupons_stack, 1, 1, 1, 1)

        coupons_sidebar = Gtk.StackSidebar()
        coupons_sidebar.set_stack(coupons_stack)
        coupons_sidebar.set_size_request(150, -1)
        coupons_section.grid.attach(coupons_sidebar, 0, 1, 1, 1)

        self.coupons_grid = Gtk.Grid()
        self.coupons_grid.set_row_spacing(10)
        coupons_stack.add_titled(self.coupons_grid, "coupons_list", _("Coupon List"))

        coupons_tree_headers = '_price', '_name', 'link', 'botid', 'token', 'assetid'
        self.coupons_tree = utils.SimpleTextTree(*coupons_tree_headers)

        price_sorter = Gtk.CustomSorter()
        price_sorter.set_sort_func(self.coupon_sorting)

        price_column = self.coupons_tree.view.get_columns()[1]
        price_column.set_sorter(price_sorter)

        self.coupons_tree.view.sort_by_column(price_column, Gtk.SortType.ASCENDING)
        self.coupons_grid.attach(self.coupons_tree, 0, 0, 4, 2)

        for index, column in enumerate(self.coupons_tree.view.get_columns()):
            column.set_resizable(True)

            if index in (0, 3, 4, 5, 6):
                column.set_visible(False)

            if index == 1:
                column.set_fixed_width(80)

            if index == 2:
                column.set_fixed_width(400)

        self.coupons_tree.view.connect("activate", self.on_coupon_double_clicked)
        self.coupons_tree.model.connect("selection-changed", self.on_tree_selection_changed)

        self.coupon_progress = Gtk.LevelBar()
        self.coupons_grid.attach(self.coupon_progress, 0, 3, 4, 1)

        fetch_coupons_button = Gtk.Button()
        fetch_coupons_button.set_margin_start(3)
        fetch_coupons_button.set_margin_end(3)
        fetch_coupons_button.set_label(_('Fetch'))
        fetch_coupons_button.connect('clicked', self.on_fetch_coupons)
        self.coupons_grid.attach(fetch_coupons_button, 0, 4, 1, 1)

        self.fetch_coupon_event = asyncio.Event()

        stop_fetching_coupons_button = Gtk.Button()
        stop_fetching_coupons_button.set_margin_start(3)
        stop_fetching_coupons_button.set_margin_end(3)
        stop_fetching_coupons_button.set_label(_('Stop fetching'))
        stop_fetching_coupons_button.connect('clicked', self.on_stop_fetching_coupons)
        self.coupons_grid.attach(stop_fetching_coupons_button, 1, 4, 1, 1)

        get_coupon_button = Gtk.Button()
        get_coupon_button.set_margin_start(3)
        get_coupon_button.set_margin_end(3)
        get_coupon_button.set_label(_('Get selected'))
        get_coupon_button.connect('clicked', self.on_coupon_action, 'get')
        self.coupons_grid.attach(get_coupon_button, 2, 4, 1, 1)

        give_coupon_button = Gtk.Button()
        give_coupon_button.set_margin_start(3)
        give_coupon_button.set_margin_end(3)
        give_coupon_button.set_label(_('Send'))
        give_coupon_button.connect('clicked', self.on_coupon_action, 'give')
        self.coupons_grid.attach(give_coupon_button, 3, 4, 1, 1)

        coupons_settings = utils.Section("coupons")
        coupons_settings.stackup_section(_("Settings"), coupons_stack)

        coupon_botids = coupons_settings.new_item("botids", _("BotIDs:"), Gtk.Entry, 0, 1)
        coupon_botids.set_placeholder_text('12345, asdfg, ...')
        coupon_botids.connect("changed", utils.on_setting_changed)

        coupon_tokens = coupons_settings.new_item("tokens", _("Tokens:"), Gtk.Entry, 0, 2)
        coupon_tokens.set_placeholder_text('12345, asdfg, ...')
        coupon_tokens.connect("changed", utils.on_setting_changed)

        coupon_botid_to_donate = coupons_settings.new_item("botid_to_donate", _("BotID To Donate:"), Gtk.Entry, 0, 3)
        coupon_botid_to_donate.connect("changed", utils.on_digit_only_setting_changed)

        coupon_token_to_donate = coupons_settings.new_item("token_to_donate", _("Token To Donate:"), Gtk.Entry, 0, 4)
        coupon_token_to_donate.connect("changed", utils.on_setting_changed)

        coupon_blacklist = coupons_settings.new_item("blacklist", _("Blacklist:"), Gtk.Entry, 0, 5)
        coupon_blacklist.connect("changed", utils.on_setting_changed)

        coupon_discount = coupons_settings.new_item(
            "minimum_discount",
            _("Minimum Discount:"),
            Gtk.DropDown,
            0, 6,
            items=config.coupon_discounts,
        )
        coupon_discount.connect("notify::selected", utils.on_dropdown_setting_changed, config.coupon_discounts)

        self.statusbar = utils.StatusBar()
        main_grid.attach(self.statusbar, 1, 3, 1, 1)

        self.connect("destroy", self.application.on_exit_activate)
        self.connect("close-request", self.application.on_exit_activate)

        self.loop = asyncio.get_event_loop()

        plugin_status_task = self.loop.create_task(self.plugin_status())
        plugin_status_task.add_done_callback(utils.safe_task_callback)

        user_info_task = self.loop.create_task(self.user_info())
        user_info_task.add_done_callback(utils.safe_task_callback)

    @property
    def theme(self) -> str:
        return config.parser.get('general', 'theme')

    async def user_info(self) -> None:
        while self.get_realized():
            account_name = config.parser.get('login', 'account_name')
            steamid_raw = config.parser.get('login', 'steamid')
            login_session = None

            with contextlib.suppress(IndexError):
                login_session = login.Login.get_session(0)

            if not login_session or not await login_session.is_logged_in():
                self.application.main_window.user_info_label.set_markup(
                    utils.markup(
                        _('Not logged in'),
                        color='darkred' if self.theme == 'light' else 'red',
                        size='small',
                    )
                )

                await asyncio.sleep(5)
                continue

            self.application.main_window.user_info_label.set_markup(
                utils.markup(
                    _('You are logged in as:\n'),
                    color='darkgreen' if self.theme == 'light' else 'green',
                    size='small',
                ) +
                utils.markup(
                    account_name,
                    color='darkblue' if self.theme == 'light' else 'blue',
                    size='small',
                ) +
                utils.markup(
                    f" {self.application.steamid.id3_string}",
                    color='grey',
                    size='small',
                )
            )

            if config.parser.get("login", "shared_secret"):
                self.steamguard_disabled.set_visible(False)
                self.steamguard_status.set_sensitive(True)
            else:
                self.steamguard_disabled.set_visible(True)
                self.steamguard_status.set_sensitive(False)
                self.steamguard_enable.set_active(False)

            if config.parser.get("login", "identity_secret"):
                self.confirmations_disabled.set_visible(False)
                self.confirmations_grid.set_sensitive(True)
            else:
                self.confirmations_disabled.set_visible(True)
                self.confirmations_grid.set_sensitive(False)
                self.confirmations_enable.set_active(False)

            self.on_stack_child_changed(self.main_tabs)
            await asyncio.sleep(30)

    async def plugin_status(self) -> None:
        while self.get_realized():
            for plugin_name in config.plugins.keys():
                if plugin_name in ["coupons", "confirmations"]:
                    if plugin_name == "confirmations":
                        enabled = config.parser.getboolean("steamguard", "enable_confirmations")
                    else:
                        enabled = config.parser.getboolean(plugin_name, "enable")

                    main = getattr(self, f'{plugin_name}_grid')
                    tree = getattr(self, f'{plugin_name}_tree')

                    if enabled:
                        tree.disabled = False
                        main.set_sensitive(True)
                    else:
                        tree.disabled = True
                        main.set_sensitive(False)
                else:
                    enabled = config.parser.getboolean(plugin_name, "enable")
                    status = getattr(self, f'{plugin_name}_status')

                    if not enabled:
                        def disabled_callback(status_) -> None:
                            status_.set_status(_("Disabled"))
                            status_.set_info("")

                        self.loop.call_later(3, disabled_callback, status)

            await asyncio.sleep(3)

    def on_fetch_coupons(self, button: Gtk.Button) -> None:
        self.fetch_coupon_event.set()

    def on_stop_fetching_coupons(self, button: Gtk.Button) -> None:
        self.fetch_coupon_event.clear()
        self.coupon_progress.set_value(0)
        self.coupon_progress.set_max_value(0)

    def on_coupon_action(self, button: Gtk.Button, action: str) -> None:
        coupon_window = coupon.CouponWindow(self, self.application, self.coupons_tree, action)
        coupon_window.present()

    def on_validate_confirmations(self, button: Gtk.Button, action: str, batch: bool = False) -> None:
        finalize_window = confirmation.FinalizeWindow(self, self.application, self.confirmations_tree, action, batch)
        finalize_window.present()

    def on_coupon_double_clicked(self, view: Gtk.ColumnView, position: int) -> None:
        row = self.coupons_tree.model.get_item(position)
        item = row.get_item()
        url = f"steam://openurl/{item.link}"
        steam_running = False

        if stlib.steamworks_available:
            with contextlib.suppress(ProcessLookupError):
                with client.SteamGameServer() as server:
                    steam_running = True

        if not steam_running:
            url = item.link

        call(f'{config.file_manager} "{url}"')

    @staticmethod
    def on_tree_selection_changed(view: Gtk.SingleSelection, position, item_count: int) -> None:
        item = view.get_selected_item()
        if parent := item.get_parent():
            view.set_selected(parent.get_position())

    @staticmethod
    def on_stack_child_changed(stack: Gtk.Stack, *args) -> None:
        section = stack.get_visible_child()

        for item in section.items:
            if item.get_name().startswith('_'):
                continue

            log.debug(f'Reading {section.get_name()}:{item.get_name()} from config file')
            item.update_values()

    @staticmethod
    def coupon_sorting(item1: utils.SimpleTextTreeItem, item2: utils.SimpleTextTreeItem, *data: Any) -> Any:
        if float(item1.price) < float(item2.price):
            return -1

        return 0 if item1.price == item2.price else 1

    def set_status(
            self,
            module: str,
            module_data: Optional[core.utils.ModuleData] = None,
            *,
            display: str = '',
            status: str = '',
            info: str = '',
            error: str = '',
            level: Tuple[int, int] = (0, 0),
            suppress_logging: bool = False,
    ) -> None:
        _status = getattr(self, f'{module}_status')

        if not module_data:
            module_data = core.utils.ModuleData(display, status, info, error, level, suppress_logging=suppress_logging)

        if module_data.display:
            if not module_data.suppress_logging:
                log.debug(f"display data: {module_data.display}")

            _status.set_display(module_data.display)
        else:
            _status.unset_display()

        if module_data.status:
            if not module_data.suppress_logging:
                log.debug(f"status data: {module_data.status}")

            _status.set_status(module_data.status)

        if module_data.info:
            if not module_data.suppress_logging:
                log.info(module_data.info)

            _status.set_info(module_data.info)

        if module_data.error:
            if not module_data.suppress_logging:
                log.error(module_data.error)

            _status.set_error(module_data.error)

        if module_data.level:
            _status.set_level(*module_data.level)

    def get_play_event(self, module: str) -> asyncio.Event:
        _status = getattr(self, f'{module}_status')
        assert isinstance(_status, utils.Status)
        return _status.play_event

    def on_login_button_clicked(self, button: Gtk.Button) -> None:
        login_window = LoginWindow(self, self.application)
        login_window.shared_secret_item.set_text('')
        login_window.identity_secret_item.set_text('')
        login_window.present()

    def on_new_authenticator_clicked(self, button: Gtk.Button) -> None:
        authenticator_window = authenticator.AuthenticatorWindow(self, self.application)
        authenticator_window.present()

        self.loop.create_task(authenticator_window.on_add_authenticator())

    def on_remove_authenticator_clicked(self, button: Gtk.Button) -> None:
        authenticator_window = authenticator.AuthenticatorWindow(self, self.application)
        authenticator_window.present()

        self.loop.create_task(authenticator_window.on_remove_authenticator())

    def on_reset_clicked(self, button: Gtk.Button) -> None:
        login_window = LoginWindow(self, self.application)
        login_window.status.info(_("Reseting... Please wait!"))
        login_window.set_deletable(False)
        login_window.user_details_section.set_visible(False)
        login_window.advanced_login.set_visible(False)
        login_window.present()

        config.config_file.unlink(missing_ok=True)

        config.parser.clear()
        config.init()

        def reset_callback() -> None:
            login_window.destroy()
            self.destroy()

        login_window.status.info(_("Successful!\nExiting..."))
        self.loop.call_later(3, reset_callback)

    def on_reset_password_clicked(self, button: Gtk.Button) -> None:
        reseting_window = utils.PopupWindowBase(self, self.application)

        reseting_status = utils.SimpleStatus()
        reseting_status.info(_("Removing saved password..."))

        reseting_window.content_grid.attach(reseting_status, 0, 0, 1, 1)
        reseting_window.set_deletable(False)
        reseting_window.present()

        config.new("login", "password", "")

        def reset_password_callback() -> None:
            reseting_window.destroy()
            self.destroy()

        reseting_status.info(_("Successful!\nExiting..."))
        self.loop.call_later(3, reset_password_callback)
